class Keywords:
    COMMAND = "collectors::keywords::command"
    PAYLOAD = "collectors::keywords::payload"


class Commands:
    EXECUTE = "collectors::commands::execute"


class Types:
    AIOHTTP = "rockingester_lib.collectors.aiohttp"
    DIRECT_POLL = "rockingester_lib.collectors.direct_poll"
